#ifdef DOORS
#error: "nostub.h" must not be included in "Doors" mode!
#endif

#ifndef NOSTUB
#define NOSTUB

#ifdef __INT_SHORT__                /* for backward compatibility */
#define int short                   /* this will supress a lot of warnings */
#else
#define int long
#endif

#define asm __asm__                 /* for ANSI mode compiling */

#ifndef OPTIMIZE_ROM_CALLS

#define _rom_call(type,args,ind) (*(type(**)args)(*(long*)0xC8+0x##ind*4))

#else

register char *__jmp_tbl asm("a5");
asm("move.l 0xC8,%a5");

#define _rom_call(type,args,ind) (*(type(**)args)(__jmp_tbl+0x##ind*4))

#endif

/* Various conditional compilations are defined below, to make an extra  */
/* support code as small as possible. If no RETURN_VALUE and SAVE_SCREEN */
/* options are defined, the code overload is only two extra bytes!       */

#ifndef RETURN_VALUE

#ifndef SAVE_SCREEN

asm("
	bset #7,(%sp)              /* The support for ~exit~ and ~atexit~   */
	jmp _main                  /* and the workaround for a bug in TIGCC */
	");

#else

asm("
	lea (%sp,-3840),%sp        /* Screen saving support */
	pea 3840                   /* (if SAVE_SCREEN is defined) */
	pea 0x4C00
	pea (%sp,8)
	move.l 0xC8,%a0
	move.l (%a0,0x26A*4),%a2
	jsr (%a2)                  /* memcpy */
	lea __RT(%pc),%a0          
	move.l %a0,%d0             
	bset #31,%d0
	move.l %d0,(%sp)
	jmp _main
__RT:
	pea 3840
	pea (%sp,12)
	pea 0x4C00
	jsr (%a2)
	lea (%sp,3860),%sp
	rts
	");

#endif

#else                              /* A RETURN_VALUE option is present */

#define __str(x) #x                /* A set of nasty hacks with preprocessor */
#define __xstr(x) __str(x)         /* based on some very specific features   */
#define __var(x) x##1
#define __xvar(x) __var(x)

#if __xvar(RETURN_VALUE)           /* Check whether RETURN_VALUE is blank */

#ifndef SAVE_SCREEN

asm("
	lea __RT(%pc),%a0          /* Support for RETURN_VALUE */
	move.l %a0,%d0
	bset #31,%d0
	move.l %d0,-(%sp)
	jmp _main
__RT:
	move.l (%sp)+,%a0
	moveq.l #4,%d0
	cmpi.w #0x21EE,(%a0)
	bne.s __A2
	moveq.l #6,%d0
__A2:
	add.l %d0,%a0
	jmp (%a0)
	");

#else

asm("
	lea (%sp,-3840),%sp        /* RETURN_VALUE, SAVE_SCREEN */
	pea 3840                
	pea 0x4C00
	pea (%sp,8)
	move.l 0xC8,%a0
	move.l (%a0,0x26A*4),%a2
	jsr (%a2)                  /* memcpy */
	lea __RT(%pc),%a0          
	move.l %a0,%d0             
	bset #31,%d0
	move.l %d0,(%sp)
	jmp _main
__RT:	pea 3840
	pea (%sp,12)
	pea 0x4C00
	jsr (%a2)
	lea (%sp,3860),%sp
	move.l (%sp)+,%a0
	moveq.l #4,%d0
	cmpi.w #0x21EE,(%a0)
	bne.s __A2
	moveq.l #6,%d0
__A2:	add.l %d0,%a0
	jmp (%a0)
	");

#endif

#else                              /* A variable is assigned to RETURN_VALUE */

#ifndef SAVE_SCREEN

asm("
	lea __RT(%pc),%a0          /* Support for RETURN_VALUE <variable> */
	move.l %a0,%d0
	bset #31,%d0
	move.l %d0,-(%sp)
	jmp _main
__RT:
	move.l 0xC8,%a0
	move.l (%a0,1060),%a1
	move.l (%a1),-(%sp)
	move.l #0x40000000,-(%sp)
	pea __VN
	move.l (%a0,536),%a0
	jsr (%a0)                  /* VarStore */
	lea (%sp,12),%sp
	rts
	dc.b 0
	.ascii \""__xstr(RETURN_VALUE)"\"
__VN:
	dc.b 0
	.even
	");

#else

asm("
	lea (%sp,-3840),%sp        /* RETURN_VALUE <variable>, SAVE_SCREEN */
	pea 3840                
	pea 0x4C00
	pea (%sp,8)
	move.l 0xC8,%a0
	move.l (%a0,0x26A*4),%a2
	jsr (%a2)                  /* memcpy */
	lea __RT(%pc),%a0          
	move.l %a0,%d0             
	bset #31,%d0
	move.l %d0,(%sp)
	jmp _main
__RT:
	pea 3840
	pea (%sp,12)
	pea 0x4C00
	jsr (%a2)
	lea (%sp,3860),%sp
	move.l 0xC8,%a0
	move.l (%a0,1060),%a1
	move.l (%a1),-(%sp)
	move.l #0x40000000,-(%sp)
	pea __VN
	move.l (%a0,536),%a0
	jsr (%a0)                  /* VarStore */
	lea (%sp,12),%sp
	rts
	dc.b 0
	.ascii \""__xstr(RETURN_VALUE)"\"
__VN:
	dc.b 0
	");

#endif

#endif

#undef __str                       /* Erase temporary macros */
#undef __xstr
#undef __var
#undef __xvar

#endif

/* Include full versions of specific patches in the program.               */
/* This is just inserted in the .s file to make tigcc.exe or ide.exe think */
/* that they have to be included.                                          */

#ifdef USE_FULL_INITS_PATCH
asm("/* Include Patch: bcopy, bzero */");
#endif

#ifdef USE_FULL_LONGMUL_PATCH
asm("/* Include Patch: __mulsi3 */");
#endif

#ifdef USE_FULL_LONGDIV_PATCH
asm("/* Include Patch: __divsi3, __modsi3, __udivsi3, __umodsi3 */");
#endif

#ifdef USE_FULL_FLOATS_PATCH
asm("/* Include Patch: __addbf3, __subbf3, __mulbf3, __divbf3, __negbf2,
           __negxf2, __fixbfsi, __fixxfsi, __fixunsbfsi, __fixunsxfsi,
           __floatsibf, __floatsixf, __nexf2, __eqxf2, __gexf2, __ltxf2,
           __gtxf2, __lexf2, __cmpbf2, __cmpxf2 */");
#endif

#define _main volatile __main__(); short _nostub; void _main

#endif
